@@ -24,8 +24,8 @@ struct policydb_compat_info {
 	int ocon_num;
 };
 
-extern struct policydb_compat_info policydb_compat[];
 extern struct policydb_compat_info *policydb_lookup_compat(int version);
+extern void __sepol_debug_printf(const char *fmt, ...);
 
 /* Reading from a policy "file". */
 static inline void *next_entry(struct policy_file * fp, size_t bytes)
@@ -55,3 +55,4 @@ static inline void *next_entry(struct policy_file * fp, size_t bytes)
 	return buffer;
 }
 
+extern int mls_enabled;
